@@ -25,7 +25,7 @@
 # OTHER DEALINGS IN THE SOFTWARE.
 from __future__ import unicode_literals
 
-version = '0.5.13'
+version = '0.5.14'
 
 import re
 import inspect
@@ -61,14 +61,15 @@ class Py3kObject(object):
         if PY3:
             return ret
         else:
-            ret.encode('utf-8')
+            return ret.encode('utf-8')
 
 from datetime import datetime
 from datetime import timedelta
 try:
-    from urllib.parse import urlsplit, parse_qs
+    from urllib.parse import urlsplit, urlunsplit, parse_qs, quote, quote_plus
 except ImportError:
-    from urlparse import urlsplit, parse_qs
+    from urlparse import urlsplit, urlunsplit, parse_qs
+    from urllib import quote, quote_plus
 
 try:
     from http.server import BaseHTTPRequestHandler
@@ -320,7 +321,6 @@ class fakesock(object):
             for matcher, value in HTTPretty._entries.items():
                 if matcher.matches(info):
                     entries = value
-                    info = matcher.info
                     break
 
             if not entries:
@@ -328,6 +328,13 @@ class fakesock(object):
                 return
 
             self._entry = matcher.get_next_entry(method)
+            # Attach more info to the entry
+            # So the callback can be more clever about what to do
+            # This does also fix the case where the callback
+            # would be handed a compiled regex as uri instead of the
+            # real uri
+            self._entry.info = info
+            self._entry.request = request
 
         def debug(*a, **kw):
             frame = inspect.stack()[0][0]
@@ -347,6 +354,8 @@ class fakesock(object):
 
         sendto = send = recvfrom_into = recv_into = recvfrom = recv = debug
 
+        def __getattr__(self, name):
+          return getattr(self.truesock, name)
 
 def fake_wrap_socket(s, *args, **kw):
     return s
@@ -464,23 +473,29 @@ class Entry(Py3kObject):
 
         self.method = method
         self.uri = uri
+        self.info = None
+        self.request = None
 
-        if isinstance(body, types.FunctionType):
-            self.body = body(method, uri, headers)
-        else:
-            self.body = body
+        self.body_is_callable = False
+        if hasattr(body,"__call__"):
+            self.callable_body = body
+            self.body = None
+            self.body_is_callable = True
+
+        self.body = body
 
         self.streaming = streaming
-        if not streaming:
+        if not streaming and not self.body_is_callable:
             self.body_length = len(self.body or '')
         else:
             self.body_length = 0
+
         self.adding_headers = adding_headers or {}
         self.forcing_headers = forcing_headers or {}
         self.status = int(status)
 
         for k, v in headers.items():
-            name = "-".join(k.split("_")).capitalize()
+            name = "-".join(k.split("_")).title()
             self.adding_headers[name] = v
 
         self.validate()
@@ -541,8 +556,11 @@ class Entry(Py3kObject):
             headers.update(self.normalize_headers(self.adding_headers))
 
         headers = self.normalize_headers(headers)
-
         status = headers.get('status', self.status)
+        if self.body_is_callable:
+            status, headers, self.body = self.callable_body(self.request,self.info.full_url(),headers)
+            headers.update({'content-length':len(self.body)})
+
         string_list = [
             'HTTP/1.1 %d %s' % (status, STATUSES[status]),
         ]
@@ -582,6 +600,13 @@ class Entry(Py3kObject):
         fk.seek(0)
 
 
+def url_fix(s, charset='utf-8'):
+    scheme, netloc, path, querystring, fragment = urlsplit(s)
+    path = quote(path, b'/%')
+    querystring = quote_plus(querystring, b':&=')
+    return urlunsplit((scheme, netloc, path, querystring, fragment))
+
+
 class URIInfo(Py3kObject):
     def __init__(self,
                  username='',
@@ -626,8 +651,16 @@ class URIInfo(Py3kObject):
         return hash(text_type(self))
 
     def __eq__(self, other):
-        self_tuple = (self.port, decode_utf8(self.hostname), decode_utf8(self.path))
-        other_tuple = (other.port, decode_utf8(other.hostname), decode_utf8(other.path))
+        self_tuple = (
+            self.port,
+            decode_utf8(self.hostname),
+            url_fix(decode_utf8(self.path)),
+        )
+        other_tuple = (
+            other.port,
+            decode_utf8(other.hostname),
+            url_fix(decode_utf8(other.path)),
+        )
         return self_tuple == other_tuple
 
     def full_url(self):
